@@ -22,6 +22,8 @@
 enum ofono_vendor {
 	OFONO_VENDOR_GENERIC = 0,
 	OFONO_VENDOR_CALYPSO,
+	OFONO_VENDOR_STE,
 	OFONO_VENDOR_QUALCOMM_MSM,
 	OFONO_VENDOR_OPTION_HSO,
+	OFONO_VENDOR_HUAWEI,
 };
