@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -26,6 +26,8 @@
 extern "C" {
 #endif
 
+#include <stdbool.h>
+
 #include <ofono/types.h>
 
 struct ofono_modem;
@@ -41,7 +43,7 @@ const char *ofono_modem_get_path(struct ofono_modem *modem);
 void ofono_modem_set_data(struct ofono_modem *modem, void *data);
 void *ofono_modem_get_data(struct ofono_modem *modem);
 
-struct ofono_modem *ofono_modem_create(const char *type);
+struct ofono_modem *ofono_modem_create(const char *name, const char *type);
 int ofono_modem_register(struct ofono_modem *modem);
 void ofono_modem_remove(struct ofono_modem *modem);
 
@@ -56,6 +58,10 @@ int ofono_modem_set_integer(struct ofono_modem *modem,
 				const char *key, int value);
 int ofono_modem_get_integer(struct ofono_modem *modem, const char *key);
 
+int ofono_modem_set_boolean(struct ofono_modem *modem,
+				const char *key, bool value);
+bool ofono_modem_get_boolean(struct ofono_modem *modem, const char *key);
+
 struct ofono_modem_driver {
 	const char *name;
 
@@ -72,8 +78,11 @@ struct ofono_modem_driver {
 	/* Power down device */
 	int (*disable)(struct ofono_modem *modem);
 
-	/* Populate the atoms supported by this device */
-	void (*populate)(struct ofono_modem *modem);
+	/* Populate the atoms available without SIM / Locked SIM */
+	void (*pre_sim)(struct ofono_modem *modem);
+
+	/* Populate the atoms that are available with SIM / Unlocked SIM*/
+	void (*post_sim)(struct ofono_modem *modem);
 };
 
 int ofono_modem_driver_register(const struct ofono_modem_driver *);
