@@ -1,7 +1,7 @@
 /*
  * oFono - GSM Telephony Stack for Linux
  *
- * Copyright (C) 2008-2009 Intel Corporation.  All rights reserved.
+ * Copyright (C) 2008-2010 Intel Corporation.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +19,10 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -55,12 +59,12 @@ static void print_scts(struct sms_scts *scts, const char *prefix)
 
 	ts = sms_scts_to_time(scts, &remote);
 
-	strftime(buf, 127, "%a, %d %b %Y %H:%M:%S %z", localtime(&ts));
+	strftime(buf, 127, "%Y-%m-%dT%H:%M:%S%z", localtime(&ts));
 	buf[127] = '\0';
 
 	g_print("local time: %s\n", buf);
 
-	strftime(buf, 127, "%a, %d %b %Y %H:%M:%S %z", &remote);
+	strftime(buf, 127, "%Y-%m-%dT%H:%M:%S%z", &remote);
 	buf[127] = '\0';
 
 	g_print("remote time: %s\n", buf);
@@ -585,7 +589,7 @@ static void test_udh_iter()
 	g_free(utf8);
 }
 
-static const char *assembly_pdu1 = "038121F350048155550119906041001222048C0500"
+static const char *assembly_pdu1 = "038121F340048155550119906041001222048C0500"
 					"031E0301041804420430043A002C002004100"
 					"43B0435043A04410430043D04340440002000"
 					"200441043B044304480430043B00200437043"
@@ -596,7 +600,7 @@ static const char *assembly_pdu1 = "038121F350048155550119906041001222048C0500"
 					"000200433043D0435";
 static int assembly_pdu_len1 = 155;
 
-static const char *assembly_pdu2 = "038121F350048155550119906041001222048C0500"
+static const char *assembly_pdu2 = "038121F340048155550119906041001222048C0500"
 					"031E03020432043E043C002E000A041D04300"
 					"43A043E043D04350446002C0020043D043500"
 					"200432002004410438043B043004450020043"
@@ -607,7 +611,7 @@ static const char *assembly_pdu2 = "038121F350048155550119906041001222048C0500"
 					"004320020043A043E";
 static int assembly_pdu_len2 = 155;
 
-static const char *assembly_pdu3 = "038121F350048155550119906041001222044A0500"
+static const char *assembly_pdu3 = "038121F340048155550119906041001222044A0500"
 					"031E0303043C043D043004420443002C00200"
 					"43F043E043704300431044B0432000A043404"
 					"3004360435002C002004470442043E0020002"
@@ -616,10 +620,10 @@ static int assembly_pdu_len3 = 89;
 
 static void test_assembly()
 {
-	unsigned char pdu[164];
+	unsigned char pdu[176];
 	long pdu_len;
 	struct sms sms;
-	struct sms_assembly *assembly = sms_assembly_new();
+	struct sms_assembly *assembly = sms_assembly_new(NULL);
 	guint16 ref;
 	guint8 max;
 	guint8 seq;
@@ -776,7 +780,7 @@ static void test_prepare_concat()
 	struct sms *sms;
 	struct sms decoded;
 	int pdu_len, tpdu_len;
-	struct sms_assembly *assembly = sms_assembly_new();
+	struct sms_assembly *assembly = sms_assembly_new(NULL);
 	guint16 ref;
 	guint8 max;
 	guint8 seq;
@@ -842,6 +846,66 @@ static void test_prepare_concat()
 	sms_assembly_free(assembly);
 }
 
+static void test_limit(gunichar uni, int target_size, gboolean use_16bit)
+{
+	char *utf8;
+	char *decoded;
+	GSList *l;
+	unsigned int i;
+	char utf8_char[6];
+	unsigned int stride;
+
+	stride = g_unichar_to_utf8(uni, utf8_char);
+
+	utf8 = g_new0(char, (target_size + 2) * stride);
+
+	for (i = 0; i < target_size * stride; i += stride)
+		memcpy(utf8 + i, utf8_char, stride);
+
+	utf8[i] = '\0';
+
+	l = sms_text_prepare(utf8, 0, use_16bit, NULL);
+
+	g_assert(l);
+	g_assert(g_slist_length(l) == 255);
+
+	decoded = sms_decode_text(l);
+	g_assert(g_utf8_strlen(decoded, -1) == target_size);
+
+	g_free(decoded);
+
+	memcpy(utf8 + i, utf8_char, stride);
+	utf8[i+stride] = '\0';
+
+	l = sms_text_prepare(utf8, 0, use_16bit, NULL);
+
+	g_assert(l == NULL);
+	g_free(utf8);
+}
+
+static void test_prepare_limits()
+{
+	gunichar ascii = 0x41;
+	gunichar ucs2 = 0x416;
+	unsigned int target_size;
+
+	/* The limit for 16 bit headers is 255 * 152 for GSM7 */
+	target_size = 255 * 152;
+	test_limit(ascii, target_size, TRUE);
+
+	/* The limit for 8 bit headers is 255 * 153 for GSM7 */
+	target_size = 255 * 153;
+	test_limit(ascii, target_size, FALSE);
+
+	/* The limit for 16 bit headers is 255 * 66 for UCS2 */
+	target_size = 255 * 66;
+	test_limit(ucs2, target_size, TRUE);
+
+	/* The limit for 8 bit headers is 255 * 67 for UCS2 */
+	target_size = 255 * 67;
+	test_limit(ucs2, target_size, FALSE);
+}
+
 static const char *cbs1 = "011000320111C2327BFC76BBCBEE46A3D168341A8D46A3D1683"
 	"41A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D168"
 	"341A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D100";
@@ -962,7 +1026,7 @@ static void test_cbs_assembly()
 	l = cbs_assembly_add_page(assembly, &dec1);
 	g_assert(l == NULL);
 
-	cbs_assembly_location_changed(assembly, TRUE, TRUE);
+	cbs_assembly_location_changed(assembly, TRUE, TRUE, TRUE);
 	g_assert(assembly->recv_cell == NULL);
 
 	dec1.update_number = 9;
@@ -1003,6 +1067,97 @@ static void test_cbs_assembly()
 	cbs_assembly_free(assembly);
 }
 
+static void test_serialize_assembly()
+{
+	unsigned char pdu[176];
+	long pdu_len;
+	struct sms sms;
+	struct sms_assembly *assembly = sms_assembly_new("1234");
+	guint16 ref;
+	guint8 max;
+	guint8 seq;
+	GSList *l;
+
+	decode_hex_own_buf(assembly_pdu1, -1, &pdu_len, 0, pdu);
+	sms_decode(pdu, pdu_len, FALSE, assembly_pdu_len1, &sms);
+
+	sms_extract_concatenation(&sms, &ref, &max, &seq);
+	l = sms_assembly_add_fragment(assembly, &sms, time(NULL),
+					&sms.deliver.oaddr, ref, max, seq);
+
+	if (g_test_verbose()) {
+		g_print("Ref: %u\n", ref);
+		g_print("Max: %u\n", max);
+		g_print("From: %s\n",
+				sms_address_to_string(&sms.deliver.oaddr));
+	}
+
+	g_assert(g_slist_length(assembly->assembly_list) == 1);
+	g_assert(l == NULL);
+
+	decode_hex_own_buf(assembly_pdu2, -1, &pdu_len, 0, pdu);
+	sms_decode(pdu, pdu_len, FALSE, assembly_pdu_len2, &sms);
+
+	sms_extract_concatenation(&sms, &ref, &max, &seq);
+	l = sms_assembly_add_fragment(assembly, &sms, time(NULL),
+					&sms.deliver.oaddr, ref, max, seq);
+	g_assert(l == NULL);
+
+	sms_assembly_free(assembly);
+
+	assembly = sms_assembly_new("1234");
+
+	decode_hex_own_buf(assembly_pdu3, -1, &pdu_len, 0, pdu);
+	sms_decode(pdu, pdu_len, FALSE, assembly_pdu_len3, &sms);
+
+	sms_extract_concatenation(&sms, &ref, &max, &seq);
+	l = sms_assembly_add_fragment(assembly, &sms, time(NULL),
+					&sms.deliver.oaddr, ref, max, seq);
+
+	g_assert(l != NULL);
+
+	sms_assembly_free(assembly);
+}
+
+static const char *ranges[] = { "1-5, 2, 3, 600, 569-900, 999",
+				"0-20, 33, 44, 50-60, 20-50, 1-5, 5, 3, 5",
+				NULL };
+static const char *inv_ranges[] = { "1-5, 3333", "1-5, afbcd", "1-5, 3-5,,",
+					"1-5, 3-5, c", NULL };
+
+static void test_range_minimizer()
+{
+	int i = 0;
+
+	while (inv_ranges[i]) {
+		GSList *l = cbs_extract_topic_ranges(inv_ranges[i]);
+
+		g_assert(l == NULL);
+		i++;
+	}
+
+	i = 0;
+
+	while (ranges[i]) {
+		GSList *r = cbs_extract_topic_ranges(ranges[i]);
+		char *rangestr;
+
+		g_assert(r != NULL);
+		i++;
+
+		rangestr = cbs_topic_ranges_to_string(r);
+
+		g_assert(rangestr);
+
+		if (g_test_verbose())
+			g_print("range: %s\n", rangestr);
+
+		g_free(rangestr);
+		g_slist_foreach(r, (GFunc)g_free, NULL);
+		g_slist_free(r);
+	}
+}
+
 int main(int argc, char **argv)
 {
 	g_test_init(&argc, &argv, NULL);
@@ -1016,10 +1171,16 @@ int main(int argc, char **argv)
 	g_test_add_func("/testsms/Test Assembly", test_assembly);
 	g_test_add_func("/testsms/Test Prepare 7Bit", test_prepare_7bit);
 	g_test_add_func("/testsms/Test Prepare Concat", test_prepare_concat);
+	g_test_add_func("/testsms/Test Prepare Limits", test_prepare_limits);
 
 	g_test_add_func("/testsms/Test CBS Encode / Decode",
 			test_cbs_encode_decode);
 	g_test_add_func("/testsms/Test CBS Assembly", test_cbs_assembly);
 
+	g_test_add_func("/testsms/Test SMS Assembly Serialize",
+			test_serialize_assembly);
+
+	g_test_add_func("/testsms/Range minimizer", test_range_minimizer);
+
 	return g_test_run();
 }
